"""diverse_seq: a tool for sampling diverse biological sequences"""

# need to import hdf5plugin here to make sure the plugin path can be
# found by h5py
import hdf5plugin  # noqa

__version__ = "2024.8.26a6"



import pathlib

import pytest


@pytest.fixture(scope="session")
def DATA_DIR():
    return pathlib.Path(__file__).parent / "data"


@pytest.fixture(scope="function")
def tmp_dir(tmp_path_factory):
    return tmp_path_factory.mktemp("dvgt")



import contextlib
import inspect
import os
import pathlib

import h5py
import hdf5plugin
import numpy
import numpy.typing
from cogent3.app.data_store import DataMember, DataStoreABC, Mode, StrOrBytes
from scitrack import get_text_hexdigest

_NOT_COMPLETED_TABLE = "not_completed"
_LOG_TABLE = "logs"
_MD5_TABLE = "md5"


_HDF5_BLOSC2_KWARGS = hdf5plugin.Blosc2(
    cname="blosclz",
    clevel=3,
    filters=hdf5plugin.Blosc2.BITSHUFFLE,
)


class HDF5DataStore(DataStoreABC):
    """Stores array data in HDF5 data sets. Associated information is
    stored as attributed on data sets.
    """

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        init_sig = inspect.signature(cls.__init__)
        bargs = init_sig.bind_partial(cls, *args, **kwargs)
        bargs.apply_defaults()
        init_vals = bargs.arguments
        init_vals.pop("self", None)
        obj._init_vals = init_vals
        return obj

    def __init__(
        self,
        source: str | pathlib.Path,
        mode: Mode = "r",
        limit: int = None,
        in_memory: bool = False,
    ) -> None:
        if in_memory:
            h5_kwargs = dict(
                driver="core",
                backing_store=False,
            )
            source = "memory"
            mode = "w"
        else:
            h5_kwargs = {}

        self._source = pathlib.Path(source)

        self._mode = Mode(mode)
        if self._mode == Mode.r and not self._source.exists():
            raise OSError(f"{self._source!s} not found")
        self._limit = limit
        self._h5_kwargs = h5_kwargs
        self._file = h5py.File(source, mode=self.mode.name, **self._h5_kwargs)
        self._is_open = True
        self._completed = []

    def __getstate__(self):
        init_vals = self._init_vals.copy()
        init_vals["mode"] = "w" if init_vals["in_memory"] else "r"
        return init_vals

    def __setstate__(self, state):
        obj = self.__class__(**state)
        self.__dict__.update(obj.__dict__)
        # because we have a __del__ method, and self attributes point to
        # attributes on obj, we need to modify obj state so that garbage
        # collection does not screw up self
        obj._is_open = False
        obj._file = None

    @property
    def source(self) -> pathlib.Path:
        """string that references connecting to data store"""
        return self._source

    @property
    def mode(self) -> Mode:
        """string that references datastore mode, override in subclass constructor"""
        return self._mode

    @property
    def limit(self):
        return self._limit

    def read(self, unique_id: str) -> numpy.ndarray:
        """reads and return array data corresponding to identifier"""
        data = self._file[unique_id]
        out = numpy.empty(len(data), dtype=numpy.uint8)
        data.read_direct(out)
        return out

    def get_attrs(self, unique_id: str) -> dict:
        """return all data set attributes connected to an identifier"""
        data = self._file[unique_id]
        return dict(data.attrs.items())

    def _write(
        self,
        *,
        subdir: str,
        unique_id: str,
        data: numpy.ndarray,
        **kwargs,
    ) -> DataMember:
        f = self._file
        path = f"{subdir}/{unique_id}" if subdir else unique_id
        dset = f.create_dataset(path, data=data, dtype="u1", **_HDF5_BLOSC2_KWARGS)

        if subdir == _LOG_TABLE:
            return None

        if subdir == _NOT_COMPLETED_TABLE:
            member = DataMember(
                data_store=self,
                unique_id=pathlib.Path(_NOT_COMPLETED_TABLE) / unique_id,
            )

        elif not subdir:
            member = DataMember(data_store=self, unique_id=unique_id)
            for key, value in kwargs.items():
                dset.attrs[key] = value

        md5 = get_text_hexdigest(data.tobytes())
        md5_dtype = h5py.string_dtype()
        f.create_dataset(f"{_MD5_TABLE}/{unique_id}", data=md5, dtype=md5_dtype)
        f.flush()
        return member

    def write(self, *, unique_id: str, data: numpy.ndarray, **kwargs) -> DataMember:
        """Writes a completed record to a dataset in the HDF5 file.

        Parameters
        ----------
        unique_id
            The unique identifier for the record. This will be used
            as the name of the dataset in the HDF5 file.
        data
            The data to be stored in the record.
        kwargs
            Any additional keyword arguments will be stored as attributes on the
            dataset. Each key-value pair in kwargs corresponds to an attribute
            name and its value.

        Returns
        -------
        DataMember
            A DataMember object representing the new record.

        Notes
        -----
        Drops any not-completed member corresponding to this identifier
        """
        member = self._write(subdir="", unique_id=unique_id, data=data, **kwargs)
        self.drop_not_completed(unique_id=unique_id)
        if member is not None:
            self._completed.append(member)
        return member

    def write_not_completed(self, *, unique_id: str, data: StrOrBytes) -> None: ...

    def write_log(self, *, unique_id: str, data: StrOrBytes) -> None: ...

    def drop_not_completed(self, *, unique_id: str | None = None) -> None: ...

    def md5(self, unique_id: str) -> str | None:
        f = self._file
        if f"md5/{unique_id}" in f:
            dset = f[f"md5/{unique_id}"]
            return dset[()].decode("utf-8")
        return None

    @property
    def completed(self) -> list[DataMember]:
        if not self._completed:
            self._completed = [
                DataMember(data_store=self, unique_id=name)
                for name in self._file.keys()
                if name not in (_LOG_TABLE, _NOT_COMPLETED_TABLE, _MD5_TABLE)
            ]
        return self._completed

    @property
    def logs(self) -> list[DataMember]:
        return []

    @property
    def not_completed(self) -> list[DataMember]:
        return []

    def __del__(self):
        self.close()

    def close(self):
        """closes the hdf5 file"""
        # hdf5 dumps content to stdout if resource already closed, so
        # we trap that here, and capture expected exceptions raised in the
        # process
        try:
            open
        except NameError:
            # builtin open() already garbage collected, so nothing to do
            return
        with open(os.devnull, "w") as devnull:
            with (
                contextlib.redirect_stderr(devnull),
                contextlib.redirect_stdout(devnull),
            ):
                with contextlib.suppress(
                    ValueError,
                    AttributeError,
                    RuntimeError,
                    PermissionError,
                ):
                    if self._is_open:
                        self._file.flush()

                with contextlib.suppress(AttributeError):
                    self._file.close()


def get_seqids_from_store(
    seq_store: str | pathlib.Path,
) -> list[str]:
    """return the list of seqids in a sequence store"""
    dstore = HDF5DataStore(seq_store, mode="r")
    return [m.unique_id for m in dstore.completed]



"""defines basic data type for storing an individual sequence record"""

import dataclasses
import functools
from collections.abc import Iterator
from math import fabs, isclose
from typing import Union

import numba
from attrs import asdict, define, field, validators
from cogent3 import get_moltype
from cogent3.app import composable
from cogent3.app import data_store as c3_data_store
from cogent3.app import typing as c3_types
from numpy import (
    array,
    errstate,
    log2,
    min_scalar_type,
    nan_to_num,
    ndarray,
    uint8,
    uint64,
    zeros,
)
from numpy import divmod as np_divmod

from diverse_seq import util as dvs_utils

NumType = Union[float, int]
PosDictType = dict[int, NumType]


@functools.singledispatch
def _gettype(name) -> type:
    try:
        return name.type
    except AttributeError:
        raise TypeError(f"type {type(name)} not supported")


@_gettype.register
def _(name: str) -> type:
    import numpy

    if name[-1].isdigit():
        return getattr(numpy, name)
    return {"int": int, "float": float}[name]


@_gettype.register
def _(name: type) -> type:
    return name


@dataclasses.dataclass(slots=True)
class lazy_kmers:
    member: c3_data_store.DataMember
    k: int
    dtype: type = uint64
    num_states: int = dataclasses.field(init=False)
    moltype: dataclasses.InitVar[str] = dataclasses.field(default="dna")

    def __post_init__(self, moltype: str):
        self.num_states = len(_get_canonical_states(moltype))

    def __array__(self):
        data = self.member.read()
        return kmer_counts(data, self.num_states, self.k, dtype=self.dtype)


@functools.singledispatch
def _make_data(data, size: int | None = None, dtype: type = None) -> ndarray:
    raise NotImplementedError


@_make_data.register
def _(data: ndarray, size: int | None = None, dtype: type = None) -> ndarray:
    return data


@_make_data.register
def _(data: lazy_kmers, size: int | None = None, dtype: type = None) -> ndarray:
    return data


@_make_data.register
def _(data: None, size: int | None = None, dtype: type = None) -> ndarray:
    return zeros(size, dtype=dtype)


@_make_data.register
def _(data: dict, size: int | None = None, dtype: type = None) -> ndarray:
    result = zeros(size, dtype=dtype)
    for i, n in data.items():
        if isclose(n, 0):
            continue
        result[i] = n
    return result


# convert this to a dataclass
# write a wrapper for the data store member so it has a __array__ method
@define(slots=True)
class vector:
    data: ndarray
    vector_length: int
    default: NumType | None = field(init=False)
    dtype: type = float
    source: str = None
    name: str = None

    def __init__(
        self,
        *,
        vector_length: int,
        data: ndarray = None,
        dtype: type = float,
        source: str = None,
        name: str = None,
    ):
        """
        Parameters
        ----------
        vector_length
            num_states**k
        data
            dict of {k-mer index: NumType}
        dtype
        """
        self.vector_length = vector_length
        dtype = _gettype(dtype)
        self.dtype = dtype

        data = _make_data(data, size=vector_length, dtype=dtype)
        self.default = dtype(0)
        self.data = data
        self.source = source
        self.name = name

    def __setitem__(self, index: int, value: NumType):
        self.data[index] = value

    def __getitem__(self, index: int) -> NumType:
        return self.data[index]

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[NumType]:
        yield from self.data

    def __getstate__(self):
        return asdict(self)

    def __setstate__(self, data):
        for k, v in data.items():
            setattr(self, k, v)
        return self

    def __sub__(self, other):
        data = self.data - other
        return self.__class__(
            data=data,
            vector_length=self.vector_length,
            dtype=self.dtype,
        )

    def __isub__(self, other):
        self.data -= other
        return self

    def __add__(self, other):
        # we are creating a new instance
        data = self.data + other
        return self.__class__(
            data=data,
            vector_length=self.vector_length,
            dtype=self.dtype,
        )

    def __iadd__(self, other):
        self.data += other
        return self

    def __truediv__(self, other):
        # we are creating a new instance
        with errstate(divide="ignore", invalid="ignore"):
            data = nan_to_num(self.data / other, nan=0.0, copy=False)
        return self.__class__(data=data, vector_length=self.vector_length, dtype=float)

    def __itruediv__(self, other):
        with errstate(divide="ignore", invalid="ignore"):
            data = nan_to_num(self.data / other, nan=0.0, copy=False)
        self.dtype = float
        self.data = data
        return self

    def sum(self):
        return self.data.sum()

    def iter_nonzero(self) -> Iterator[NumType]:
        yield from (v for v in self.data if v)

    @property
    def entropy(self):
        non_zero = self.data[self.data > 0]
        kfreqs = non_zero if self.dtype == float else non_zero / non_zero.sum()
        # taking absolute value due to precision issues
        return fabs(-(kfreqs * log2(kfreqs)).sum())

    def __array__(self):
        if not isinstance(self.data, ndarray):
            self.data = array(self.data)
        return self.data


@numba.jit(nopython=True)
def coord_conversion_coeffs(num_states, k):  # pragma: no cover
    """coefficients for multi-dimensional coordinate conversion into 1D index"""
    return array([num_states ** (i - 1) for i in range(k, 0, -1)])


@numba.jit(nopython=True)
def coord_to_index(coord, coeffs):  # pragma: no cover
    """converts a multi-dimensional coordinate into a 1D index"""
    return (coord * coeffs).sum()


@numba.jit(nopython=True)
def index_to_coord(index, coeffs):  # pragma: no cover
    """converts a 1D index into a multi-dimensional coordinate"""
    ndim = len(coeffs)
    coord = zeros(ndim, dtype=uint64)
    remainder = index
    for i in range(ndim):
        n, remainder = np_divmod(remainder, coeffs[i])
        coord[i] = n
    return coord


def indices_to_seqs(indices: ndarray, states: bytes, k: int) -> list[str]:
    """convert indices from k-dim into sequence

        Parameters
    ----------
    indices
        array of (len(states), k) dimensioned indices
    states
        the ordered characters, e.g. b"TCAG"
    """
    arr = indices_to_bytes(indices, states, k)
    return [bytearray(kmer).decode("utf8") for kmer in arr]


@numba.jit(nopython=True)
def indices_to_bytes(
    indices: ndarray,
    states: bytes,
    k: int,
) -> ndarray:  # pragma: no cover
    """convert indices from k-dim into bytes

        Parameters
    ----------
    indices
        array of (len(states), k) dimensioned indices
    states
        the ordered characters, e.g. b"TCAG"
    k
        dimensions

    Raises
    -----
    IndexError if an index not in states

    Returns
    -------
    uint8 array.

    Notes
    -----
    Use indices_to_seqs to have the results returned as strings
    """
    result = zeros((len(indices), k), dtype=uint8)
    coeffs = coord_conversion_coeffs(len(states), k)
    num_states = len(states)
    for i in range(len(indices)):
        index = indices[i]
        coord = index_to_coord(index, coeffs)
        for j in range(k):
            if coord[j] >= num_states:
                raise IndexError("index out of character range")
            result[i][j] = states[coord[j]]

    return result


@numba.jit(nopython=True)
def kmer_counts(
    seq: ndarray,
    num_states: int,
    k: int,
    dtype=uint64,
) -> ndarray:  # pragma: no cover
    """return freqs of valid k-mers using 1D indices

    Parameters
    ----------
    seq
        numpy array of uint, assumed that canonical characters have
        sequential indexes which are all < num_states
    result
        array to be written into
    num_states
        defines range of possible ints at a position
    k
        k-mer size

    Returns
    -------
    result
    """
    coeffs = coord_conversion_coeffs(num_states, k)
    kfreqs = zeros(num_states**k, dtype=dtype)
    skip_until = 0
    for i in range(k):
        if seq[i] >= num_states:
            skip_until = i + 1

    for i in range(len(seq) - k + 1):
        if seq[i + k - 1] >= num_states:
            skip_until = i + k

        if i < skip_until:
            continue

        kmer = seq[i : i + k]
        index = (kmer * coeffs).sum()
        kfreqs[index] += 1

    return kfreqs


def _gt_zero(instance, attribute, value):
    if value <= 0:
        raise ValueError(f"must be > 0, not {value}")


@functools.singledispatch
def _make_kcounts(data) -> vector:
    raise TypeError(f"type {type(data)} not supported")


@_make_kcounts.register
def _(data: ndarray):
    nonzero = {i: v for i, v in enumerate(data.tolist()) if v}
    return vector(
        vector_length=len(data),
        data=nonzero,
        dtype=_gettype(data.dtype.name),
    )


@_make_kcounts.register
def _(data: vector):
    return data


@dataclasses.dataclass(frozen=True)
class SeqArray:
    """A SeqArray stores an array of indices that map to the canonical characters
    of the moltype of the original sequence. Use rep_seq.util.arr2str() to
    recapitulate the original sequence."""

    seqid: str
    data: ndarray
    moltype: str
    source: str = None

    def __len__(self):
        return len(self.data)


@define(slots=True, order=True, hash=True)
class KmerSeq:
    """representation of a single sequence as kmer counts"""

    kcounts: vector = field(eq=False, converter=_make_kcounts)
    name: str = field(validator=validators.instance_of(str), eq=True)
    delta_jsd: float = field(
        init=False,
        validator=validators.instance_of(float),
        default=0.0,
        eq=False,
    )

    @property
    def size(self):
        return self.kcounts.vector_length

    @functools.cached_property
    def entropy(self):
        return self.kfreqs.entropy

    @functools.cached_property
    def kfreqs(self):
        kcounts = array(self.kcounts)
        kcounts = kcounts.astype(float)
        kfreqs = kcounts / kcounts.sum()
        return vector(
            data=kfreqs,
            vector_length=len(kfreqs),
            dtype=float,
        )


@composable.define_app
class seq_to_seqarray:
    """Convert a cogent3 Sequence to a SeqArray"""

    def __init__(
        self,
        moltype: str = "dna",
    ):
        self.moltype = moltype
        self.str2arr = dvs_utils.str2arr(moltype=self.moltype)

    def main(self, seq: c3_types.SeqType) -> SeqArray:
        return SeqArray(
            seqid=seq.name,
            data=self.str2arr(str(seq)),
            moltype=self.moltype,
            source=seq.info.source or seq.name,
        )


@composable.define_app
class seqarray_to_kmerseq:
    """converts a sequence numpy array to a KmerSeq"""

    def __init__(self, k: int, moltype: str):
        """
        Parameters
        ----------
        k
            size of k-mers
        moltype
            cogent3 molecular type
        """
        self.moltype = moltype
        self.k = k
        self.canonical = _get_canonical_states(moltype)
        self.dtype = min_scalar_type(len(self.canonical) ** self.k)

    def main(self, seq: SeqArray) -> KmerSeq:
        kwargs = dict(
            vector_length=len(self.canonical) ** self.k,
            dtype=int,
            source=seq.source,
            name=seq.seqid,
        )
        counts = kmer_counts(seq.data, len(self.canonical), self.k, dtype=self.dtype)
        kwargs["data"] = counts

        return KmerSeq(
            kcounts=vector(**kwargs),
            name=kwargs["name"],
        )


@composable.define_app
class member_to_kmerseq:
    """creates a KmerSeq from a HDF5 datastore member"""

    def __init__(self, k: int, moltype: str):
        """
        Parameters
        ----------
        k
            size of k-mers
        moltype
            cogent3 molecular type
        """
        self.moltype = moltype
        self.k = k
        self.num_states = len(_get_canonical_states(moltype))
        self.dtype = min_scalar_type(self.num_states**k)

    def main(self, member: c3_data_store.DataMember) -> KmerSeq:
        vec = lazy_kmers(
            member=member,
            k=self.k,
            moltype=self.moltype,
            dtype=self.dtype,
        )
        kwargs = dict(
            vector_length=vec.num_states,
            dtype=self.dtype,
            source=member.data_store.source,
            name=member.unique_id,
            data=vec,
        )

        return KmerSeq(
            kcounts=vector(**kwargs),
            name=kwargs["name"],
        )


@functools.cache
def _get_canonical_states(moltype: str) -> bytes:
    moltype = get_moltype(moltype)
    canonical = list(moltype.alphabet)
    v = moltype.alphabet.to_indices(canonical)
    if not (0 <= min(v) < max(v) < len(canonical)):
        raise ValueError(f"indices of canonical states {canonical} not sequential {v}")
    return "".join(canonical).encode("utf8")



import pytest
from numpy import array
from numpy.testing import assert_allclose

from diverse_seq.distance import _intersect_union, jaccard


def _make_sample(type_):
    return type_([0, 1, 3]), type_([1, 4, 5])


@pytest.mark.parametrize("rec1,rec2", (_make_sample(set), _make_sample(array)))
def test_intersect_union(rec1, rec2):
    i, u = _intersect_union(rec1, rec2)
    assert i == 1 and u == 5


@pytest.mark.parametrize(
    "rec1,rec2,exp",
    (
        [{1, 2}, {3, 4}, 1],
        [set(), {1}, 1],
        [{1}, set(), 1],
        [{1, 2}, {1, 2}, 0],
        [{1, 2}, {1, 3}, 2 / 3],
    ),
)
def test_jaccard(rec1, rec2, exp):
    assert_allclose(jaccard(rec1, rec2), exp)



from functools import singledispatch

from numpy import intersect1d, ndarray, union1d


@singledispatch
def _intersect_union(rec1, rec2) -> tuple[int, int]:
    raise NotImplementedError


@_intersect_union.register(set)
def _(rec1: set, rec2: set) -> tuple[int, int]:
    intersect = rec1 & rec2
    union = rec1 | rec2
    return len(intersect), len(union)


@_intersect_union.register(ndarray)
def _(rec1: ndarray, rec2: ndarray) -> tuple[int, int]:
    intersect = intersect1d(rec1, rec2, assume_unique=True)
    union = union1d(rec1.data, rec2.data)
    return len(intersect), len(union)


def jaccard(rec1, rec2) -> float:
    """returns the Jaccard distance between rec1 and rec2"""
    i, u = _intersect_union(rec1, rec2)
    return 1 - (i / u)



"""summed records contain multiple SeqRecord instances. This container class
efficiently computes jsd

Convert a sequence collection into a SummedRecords instance, which supports
the following applications

Find N
-------
identify the n-most diverse sequences

Most divergent
--------------
identify the set of sequences that maximise delta-JSD

SummedRecords is the container that simplifies these applications
"""

import functools
import itertools
import pathlib
import sys
import typing
from math import fsum

import numpy
from attrs import define, field
from cogent3 import make_table
from cogent3.app import typing as c3_types
from cogent3.app.composable import NotCompleted, define_app
from numpy import isclose as np_isclose
from rich import progress as rich_progress

from diverse_seq import data_store as dvs_data_store
from diverse_seq import util as dvs_util
from diverse_seq.record import (
    KmerSeq,
    member_to_kmerseq,
    seq_to_seqarray,
    seqarray_to_kmerseq,
    vector,
)

# needs a jsd method for a new sequence
# needs summed entropy scores
# needs combined array of summed freqs to allow efficient calculation of entropy
# on creation, computes total JSD, then delta_jsd (contribution of each sequence)
# seq records sorted by the latter; stored summed quantities for n and n-1 (with
# the least contributor omitted)


@functools.singledispatch
def _jsd(summed_freqs: vector | numpy.ndarray, summed_entropy: float, n: int) -> float:
    raise NotImplementedError


@_jsd.register
def _(summed_freqs: vector, summed_entropy: float, n: int):
    kfreqs = summed_freqs / n
    entropy = summed_entropy / n
    return kfreqs.entropy - entropy


@_jsd.register
def _(summed_freqs: numpy.ndarray, summed_entropy: float, n: int):
    kfreqs = summed_freqs / n
    entropy = summed_entropy / n
    kfreqs = kfreqs[~np_isclose(kfreqs, 0)]
    ke = -(kfreqs * numpy.log2(kfreqs)).sum()
    return ke - entropy


def _summed_stats(records: list[KmerSeq]) -> tuple[vector, float]:
    # takes series of records and sums quantitative parts
    sv = records[0].kfreqs
    vec = numpy.zeros(len(sv), dtype=float)
    entropies = []
    for record in records:
        vec += record.kfreqs
        entropies.append(record.entropy)

    vec = vector(data=vec, vector_length=len(vec), dtype=float)
    return vec, fsum(entropies)


def _delta_jsd(
    total_kfreqs: vector,
    total_entropies: float,
    records: list[KmerSeq],
) -> list[KmerSeq]:
    """measures contribution of each record to the total JSD"""
    n = len(records)
    total_jsd = _jsd(total_kfreqs, total_entropies, n)
    total_kfreqs = total_kfreqs.data
    result = []
    for record in records:
        summed_kfreqs = total_kfreqs - record.kfreqs.data
        summed_entropies = total_entropies - record.entropy
        jsd = _jsd(summed_kfreqs, summed_entropies, n - 1)
        record.delta_jsd = total_jsd - jsd
        if numpy.isnan(record.delta_jsd):
            print(f"{record.name!r} had a nan")
            sys.exit(1)
        result.append(record)
    return result


def _check_integrity(instance, attribute, records: list[KmerSeq]):
    last = records[0]
    for r in records[1:]:
        if r.delta_jsd < last.delta_jsd:
            raise RuntimeError


@define
class SummedRecords:
    """use the from_records clas method to construct the instance"""

    # we need the records to have been sorted by delta_jsd
    # following check is in place for now until fully tested
    # TODO delete when convinced no longer required

    records: list[KmerSeq] = field(validator=_check_integrity)
    summed_kfreqs: vector
    summed_entropies: float
    total_jsd: float
    size: int = field(init=False)
    record_names: set = field(init=False)
    lowest: KmerSeq = field(init=False)
    _stats: dvs_util.summary_stats = field(init=False)

    def __init__(
        self,
        records: list[KmerSeq],
        summed_kfreqs: vector,
        summed_entropies: float,
        total_jsd: float,
    ):
        self.record_names = {r.name for r in records}
        self.total_jsd = total_jsd
        self.lowest = records[0]
        self.records = records[1:]
        self.size = len(records)
        # NOTE we exclude lowest record from freqs and entropy
        self.summed_kfreqs = summed_kfreqs - self.lowest.kfreqs
        self.summed_entropies = summed_entropies - self.lowest.entropy
        self._stats = dvs_util.summary_stats(
            numpy.array([r.delta_jsd for r in records]),
        )

    @classmethod
    def from_records(cls, records: list[:KmerSeq]):
        size = len(records)
        summed_kfreqs, summed_entropies = _summed_stats(records)
        total_jsd = _jsd(summed_kfreqs, summed_entropies, size)
        records = sorted(
            _delta_jsd(summed_kfreqs, summed_entropies, records),
            key=lambda x: x.delta_jsd,
        )
        return cls(records, summed_kfreqs, summed_entropies, total_jsd)

    def _make_new(
        self,
        records: list[:KmerSeq],
        summed_kfreqs: vector,
        summed_entropies: float,
    ):
        """summed are totals from all records"""
        size = len(records)
        total_jsd = _jsd(summed_kfreqs, summed_entropies, size)
        records = sorted(
            _delta_jsd(
                summed_kfreqs,
                summed_entropies,
                records,
            ),
            key=lambda x: x.delta_jsd,
        )
        return self.__class__(records, summed_kfreqs, summed_entropies, total_jsd)

    def __contains__(self, item: KmerSeq):
        return item.name in self.record_names

    def __add__(self, other: KmerSeq):
        assert other not in self
        summed_kfreqs = self.summed_kfreqs + self.lowest.kfreqs + other.kfreqs
        summed_entropies = self.summed_entropies + self.lowest.entropy + other.entropy
        return self._make_new(
            [self.lowest, other] + list(self.records),
            summed_kfreqs,
            summed_entropies,
        )

    def __sub__(self, other: KmerSeq):
        if other not in self:
            raise ValueError(
                f"cannot subtract record {other.name!r}, not present in self",
            )
        records = [r for r in self.records + [self.lowest] if r is not other]

        summed_kfreqs = self.summed_kfreqs + self.lowest.kfreqs - other.kfreqs
        summed_entropies = self.summed_entropies + self.lowest.entropy - other.entropy
        return self._make_new(records, summed_kfreqs, summed_entropies)

    def iter_record_names(self):
        yield from self.record_names

    def increases_jsd(self, record: KmerSeq) -> bool:
        # whether total JSD increases when record is used
        j = _jsd(
            self.summed_kfreqs + record.kfreqs,
            self.summed_entropies + record.entropy,
            self.size,
        )
        return self.total_jsd < j

    @property
    def mean_jsd(self):
        return self.total_jsd / self.size

    @property
    def mean_delta_jsd(self):
        """mean of delta_jsd"""
        return self._stats.mean

    @property
    def std_delta_jsd(self):
        """unbiased standard deviation of delta_jsd"""
        return self._stats.std

    @property
    def cov_delta_jsd(self):
        """coeff of variation of delta_jsd"""
        return self._stats.cov

    def replaced_lowest(self, other: KmerSeq):
        """returns new instance with other instead of lowest"""
        summed_kfreqs = self.summed_kfreqs + other.kfreqs
        summed_entropies = self.summed_entropies + other.entropy
        return self._make_new([other] + self.records, summed_kfreqs, summed_entropies)

    def to_table(self):
        names, deltas = list(
            zip(
                *[(r.name, r.delta_jsd) for r in [self.lowest] + self.records],
                strict=False,
            ),
        )
        return make_table(data={"names": names, "delta_jsd": deltas})

    def all_records(self):
        """returns all records in order of delta_jsd"""
        return [self.lowest] + self.records


def _get_stat_attribute(stat: str) -> str:
    if stat not in ("stdev", "cov"):
        raise ValueError(f"unknown value of stat {stat!r}")

    return {"stdev": "std_delta_jsd", "cov": "cov_delta_jsd"}[stat]


def max_divergent(
    records: list[KmerSeq],
    min_size: int = 2,
    max_size: int | None = None,
    stat: str = "stdev",
    max_set: bool = True,
    verbose: bool = False,
) -> SummedRecords:
    """returns SummedRecords that maximises stat

    Parameters
    ----------
    records
        list of SeqRecord instances
    min_size
        starting size of SummedRecords
    max_size
        defines upper limit of SummedRecords size
    stat
        either stdev or cov, which represent the statistics
        std(delta_jsd) and cov(delta_jsd) respectively
    max_set
        postprocess to identify subset that maximises stat

    Notes
    -----
    This is sensitive to the order of records.
    """
    max_size = max_size or len(records)
    sr = SummedRecords.from_records(records[:min_size])

    attr = _get_stat_attribute(stat)

    if len(records) <= min_size:
        return sr

    series = rich_progress.track(records, transient=True) if verbose else records
    for r in series:
        if r in sr:
            # already a member of the divergent set
            continue

        if not sr.increases_jsd(r):
            # does not increase total JSD
            continue

        nsr = sr + r
        sr = nsr if getattr(nsr, attr) > getattr(sr, attr) else sr.replaced_lowest(r)
        if sr.size > max_size:
            sr = SummedRecords.from_records(sr.records)

    if max_set:
        app = dvs_final_max(stat=stat, min_size=min_size, verbose=verbose)  # pylint: disable=no-value-for-parameter
        sr = app([sr])
    elif verbose:
        num_neg = sum(r.delta_jsd < 0 for r in [sr.lowest] + sr.records)
        print(f"Records with delta_jsd < 0 is {num_neg}")
    return sr


@define_app
def dvs_final_max(
    summed: list[SummedRecords],
    *,
    stat: str,
    min_size: int,
    verbose: bool,
) -> SummedRecords:
    """returns the set that maximises stat

    Parameters
    ----------
    sr
        SummedRecords instance
    stat
        name of a SummedRecords attribute that returns the statistic
    min_size
        the minimum size of the set
    verbose
        display extra information
    """
    if len(summed) > 1:
        records = list(itertools.chain.from_iterable(sr.all_records() for sr in summed))
        numpy.random.shuffle(records)
        sr = SummedRecords.from_records(records)
    else:
        sr = summed[0]

    if sr.size == min_size:
        return sr

    stat = _get_stat_attribute(stat)

    results = {getattr(sr, stat): sr}
    orig_size = sr.size
    orig_stat = getattr(sr, stat)

    while sr.size > min_size:
        sr = sr.from_records(sr.records)
        results[getattr(sr, stat)] = sr

    sr = results[max(results)]

    if verbose:
        print(
            f"Reduced size from {orig_size} to {sr.size}",
            f" Increased {stat} from {orig_stat:.3f} to {getattr(sr, stat):.3f}",
            sep="\n",
        )
    return sr


def most_divergent(
    records: list[KmerSeq],
    size: int,
    verbose: bool = False,
    show_progress: bool = False,
) -> SummedRecords:
    """returns size most divergent records

    Parameters
    ----------
    records
        list of SeqRecord instances
    size
        starting size of SummedRecords
    show_progress
        display progress bar
    """
    size = size or 2
    sr = SummedRecords.from_records(records[:size])

    if len(records) <= size:
        return sr

    series = rich_progress.track(records) if show_progress else records
    for r in series:
        if r in sr:
            continue

        if not sr.increases_jsd(r):
            continue

        sr = sr.replaced_lowest(r)

    if verbose:
        num_neg = sum(r.delta_jsd < 0 for r in [sr.lowest] + sr.records)
        print(f"number negative is {num_neg}")
    return sr


@define_app
class dvs_max:
    """return the maximally divergent sequences"""

    def __init__(
        self,
        *,
        seq_store: str | pathlib.Path,
        k: int = 3,
        min_size: int = 7,
        max_size: int = None,
        stat: str = "stdev",
        limit: int = None,
        verbose=0,
    ) -> None:
        """

        Parameters
        ----------
        seq_store
            path to divergent sequence store
        k
            k-mer size
        min_size
            the minimum number of sequences to be included in the divergent set
        max_size
            the maximum number of sequences to be included in the divergent set
        stat
            the stat to use for selecting whether to include a sequence
        limit
            limit number of sequence records to process
        stat
            the statistic to use for optimising, by default "cov_delta_jsd"
        verbose
            extra info display
        """
        self._seq_store = seq_store
        self._k = k
        self._limit = limit
        self._max_size = max_size
        self._min_size = min_size
        self._stat = stat
        self._verbose = verbose

    def main(self, seq_names: list[str]) -> SummedRecords:
        records = records_from_seq_store(
            seq_store=self._seq_store,
            seq_names=seq_names,
            limit=self._limit,
            k=self._k,
        )
        # TODO: add ability to set random number seed
        numpy.random.shuffle(records)
        return max_divergent(
            records=records,
            min_size=self._min_size,
            max_size=self._max_size,
            stat=self._stat,
            verbose=self._verbose > 0,
            max_set=False,
        )


def records_from_seq_store(
    *,
    seq_store: str | pathlib.Path,
    seq_names: list[str],
    k: int,
    moltype: str = "dna",
    limit: int | None,
) -> list[KmerSeq]:
    """converts sequences in seq_store into SeqRecord instances

    Parameters
    ----------
    seq_store
        path to divergent sequence store
    seq_names
        list of names that are members of the seq_store
    moltype
        the expected molecular type
    limit
        limit number of sequence records to process
    k
        k-mer size

    Returns
    -------
        sequences are converted into vector of k-mer counts
    """
    dstore = dvs_data_store.HDF5DataStore(seq_store, mode="r")
    make_record = member_to_kmerseq(k=k, moltype=moltype)
    records = [make_record(m) for m in dstore.completed if m.unique_id in seq_names]  # pylint: disable=not-callable
    records = records[:limit] if limit else records
    for record in records:
        if not record:
            print(record)
            sys.exit(1)
    return records


@define_app
class dvs_nmost:
    """return the N most divergent sequences"""

    def __init__(
        self,
        *,
        seq_store: str | pathlib.Path,
        k: int = 3,
        limit: int = None,
        n: int,
        verbose=0,
    ) -> None:
        """

        Parameters
        ----------
        seq_store
            path to divergent sequence store
        n
            the number of divergent sequences
        k
            k-mer size
        limit
            limit number of sequence records to process
        verbose
            extra info display
        """
        self._seq_store = seq_store
        self._k = k
        self._limit = limit
        self._max_size = n
        self._verbose = verbose

    def main(self, seq_names: list[str]) -> SummedRecords:
        records = records_from_seq_store(
            seq_store=self._seq_store,
            seq_names=seq_names,
            limit=self._limit,
            k=self._k,
        )
        # TODO: add ability to set random number seed
        numpy.random.shuffle(records)
        return most_divergent(records, size=self._max_size, verbose=self._verbose > 0)


@define_app
def dvs_final_nmost(summed: list[SummedRecords]) -> SummedRecords:
    """selects the best n records from a list of SummedRecords

    Notes
    -----
    Useful for aggregating results from multiple runs.
    """
    size = max(sr.size for sr in summed)
    records = list(itertools.chain.from_iterable(sr.all_records() for sr in summed))
    numpy.random.shuffle(records)
    return most_divergent(records, size=size, verbose=False, show_progress=False)


def apply_app(
    *,
    app: dvs_max,
    seqids: list[str],
    numprocs: int,
    verbose: bool,
    hide_progress: bool = False,
    finalise: typing.Callable[[list[SummedRecords]], SummedRecords],
) -> SummedRecords:
    """applies the app to seqids, polishing the selected set with finalise"""
    app_name = app.__class__.__name__
    with dvs_util.keep_running():
        if numprocs > 1 and len(seqids) > numprocs:
            seqids = list(dvs_util.chunked(seqids, numprocs, verbose=verbose))
        else:
            seqids = [seqids]

        with rich_progress.Progress(
            rich_progress.TextColumn("{task.description}"),
            rich_progress.BarColumn(),
            rich_progress.TaskProgressColumn(),
            rich_progress.TimeRemainingColumn(),
            rich_progress.TimeElapsedColumn(),
            disable=hide_progress,
        ) as progress:
            select = progress.add_task(
                description=f"[blue]Selection with {app_name!r}",
                total=len(seqids),
            )
            result = []
            for r in app.as_completed(
                seqids,
                parallel=numprocs > 1,
                par_kw=dict(max_workers=numprocs),
                show_progress=True,
            ):
                if not r:
                    dvs_util.print_colour(r, style="red")
                result.append(r.obj)
                progress.update(select, advance=1, refresh=True)

        dvs_util.print_colour(f"Merging results from {len(seqids)} runs...", "blue")

        result = finalise(result)

        if isinstance(result, NotCompleted):
            dvs_util.print_colour(f"{result.type}: {result.message}", "red")
            sys.exit(1)

    return result


# single apps that combine multiple apps to work off a sequence collection
# app convert seq coll to SequenceArray and converts those into KmerSeq instances


@define_app
class dvs_select_max:
    """selects the maximally divergent seqs from a sequence collection"""

    def __init__(
        self,
        min_size: int = 5,
        max_size: int = 30,
        stat: str = "stdev",
        moltype: str = "dna",
        include: list[str] | str | None = None,
        k: int = 6,
        seed: int | None = None,
    ) -> None:
        """
        Parameters
        ----------
        min_size
            minimum size of the divergent set
        max_size
            the maximum size of the divergent set
        stat
            either stdev or cov, which represent the statistics
            std(delta_jsd) and cov(delta_jsd) respectively
        moltype
            molecular type of the sequences
        include
            sequence names to include in the final result
        k
            k-mer size
        seed
            random number seed

        Notes
        -----
        If called with an alignment, the ungapped sequences are used.
        The order of the sequences is randomised. If include is not None, the
        named sequences are added to the final result.
        """
        self._s2k = seq_to_seqarray(moltype=moltype) + seqarray_to_kmerseq(
            k=k,
            moltype=moltype,
        )
        self._min_size = min_size
        self._max_size = max_size
        self._stat = stat
        self._rng = numpy.random.default_rng(seed)
        self._include = [include] if isinstance(include, str) else include

    def main(self, seqs: c3_types.SeqsCollectionType) -> c3_types.SeqsCollectionType:
        records = [self._s2k(seqs.get_seq(name)) for name in seqs.names]
        self._rng.shuffle(records)
        for record in records:
            if not record:
                print(record)
                return None
        result = max_divergent(
            records=records,
            min_size=self._min_size,
            max_size=self._max_size,
            stat=self._stat,
        )
        selected = set(result.record_names) | set(self._include or [])
        return seqs.take_seqs(selected)


@define_app
class dvs_select_nmost:
    """selects the n-most diverse seqs from a sequence collection"""

    def __init__(
        self,
        n: int = 10,
        moltype: str = "dna",
        include: list[str] | str | None = None,
        k: int = 6,
        seed: int | None = None,
    ) -> None:
        """
        Parameters
        ----------
        n
            the number of divergent sequences
        moltype
            molecular type of the sequences
        k
            k-mer size
        include
            sequence names to include in the final result
        seed
            random number seed

        Notes
        -----
        If called with an alignment, the ungapped sequences are used.
        The order of the sequences is randomised. If include is not None, the
        named sequences are added to the final result.
        """
        self._s2k = seq_to_seqarray(moltype=moltype) + seqarray_to_kmerseq(
            k=k,
            moltype=moltype,
        )
        self._n = n
        self._moltype = moltype
        self._rng = numpy.random.default_rng(seed)
        self._include = [include] if isinstance(include, str) else include

    def main(self, seqs: c3_types.SeqsCollectionType) -> c3_types.SeqsCollectionType:
        records = [self._s2k(seqs.get_seq(name)) for name in seqs.names]
        self._rng.shuffle(records)
        result = most_divergent(
            records=records,
            size=self._n,
        )
        selected = set(result.record_names) | set(self._include or [])
        return seqs.take_seqs(selected)



import pathlib

import pytest
from cogent3 import load_unaligned_seqs, make_unaligned_seqs
from numpy.testing import assert_array_equal

from diverse_seq import data_store
from diverse_seq import io as dvs_io
from diverse_seq import record as dvs_record
from diverse_seq.util import str2arr

DATADIR = pathlib.Path(__file__).parent / "data"


@pytest.fixture(scope="function")
def brca1_seqs():
    return load_unaligned_seqs(DATADIR / "brca1.fasta", moltype="dna").degap()


@pytest.fixture(scope="function")
def brca1_5(brca1_seqs):
    seqs = brca1_seqs.take_seqs(["Cat", "Dog", "Wombat", "Horse", "Rat"])
    return make_unaligned_seqs(
        data={s.name: str(s[:20]) for s in seqs.seqs},
        moltype="dna",
    )


@pytest.fixture(scope="function")
def brca1_dstore(tmp_path):
    dstore_maker = dvs_io.dvs_file_to_dir(tmp_path / "brca1_dstore")
    return dstore_maker(DATADIR / "brca1.fasta")  # pylint: disable=not-callable


@pytest.fixture(scope="function")
def brca1_5_dstore(tmp_path, brca1_5):
    fasta_path = tmp_path / "brca1_5.fasta"
    brca1_5.write(fasta_path)
    dstore_maker = dvs_io.dvs_file_to_dir(tmp_path / "brca1_5_dstore")
    return dstore_maker(fasta_path)  # pylint: disable=not-callable


@pytest.fixture(scope="function")
def hdf5_dstore_path(tmp_path):
    return tmp_path / "hdf5_dstore"


def test_dvs_file_to_dir(brca1_5_dstore, brca1_5):
    # directory contains the same number of sequences as the input file
    assert brca1_5.num_seqs == len(brca1_5_dstore.completed)


@pytest.mark.parametrize("parallel", (False, True))
def test_prep_pipeline(brca1_5, brca1_5_dstore, hdf5_dstore_path, parallel):
    # initialise and apply pipeline
    dstore = data_store.HDF5DataStore(source=hdf5_dstore_path, mode="w")
    prep = dvs_io.dvs_load_seqs(moltype="dna") + dvs_io.dvs_write_seqs(
        data_store=dstore,
    )
    result = prep.apply_to(brca1_5_dstore, parallel=parallel)

    # output datastore contains same number of records as seqs in orig file
    assert brca1_5.num_seqs == len(result.completed)

    # check the sequence data matches
    seq_data = result.read("Cat")
    str_to_array = str2arr(moltype="dna")
    orig_seq_data = str_to_array(str(brca1_5.get_seq("Cat")))  # pylint: disable=not-callable
    assert_array_equal(seq_data, orig_seq_data)


def test_get_seqids(DATA_DIR):
    fasta_path = DATA_DIR / "brca1.fasta"
    expect = set(load_unaligned_seqs(fasta_path).names)
    store_path = DATA_DIR / "brca1.dvseqs"

    got = data_store.get_seqids_from_store(store_path)

    assert set(got) == expect


def test_in_memory(brca1_5):
    dstore = data_store.HDF5DataStore(source=":memory:", mode="w", in_memory=True)
    writer = dvs_io.dvs_write_seqs(
        data_store=dstore,
    )
    prep = dvs_record.seq_to_seqarray(moltype="dna") + writer
    prep.apply_to(brca1_5.seqs, id_from_source=lambda x: x.name, logger=False)
    assert len(dstore) == brca1_5.num_seqs
    assert dstore.read("Cat").shape == (20,)



import contextlib
import functools
import math
import pathlib
import re

import numpy
from cogent3 import get_moltype
from cogent3.app import composable
from rich import text as rich_text

try:
    from wakepy.keep import running as keep_running

    # trap flaky behaviour on linux
    with keep_running():
        ...

except (NotImplementedError, ImportError):

    @contextlib.contextmanager
    def keep_running(*args, **kwargs):
        yield


@composable.define_app
class str2arr:
    """convert string to array of uint8"""

    def __init__(self, moltype: str = "dna", max_length=None):
        moltype = get_moltype(moltype)
        self.canonical = "".join(moltype)
        self.max_length = max_length
        extended = "".join(list(moltype.alphabets.degen))
        self.translation = b"".maketrans(
            extended.encode("utf8"),
            "".join(chr(i) for i in range(len(extended))).encode("utf8"),
        )

    def main(self, data: str) -> numpy.ndarray:
        if self.max_length:
            data = data[: self.max_length]

        b = data.encode("utf8").translate(self.translation)
        return numpy.array(memoryview(bytearray(b)), dtype=numpy.uint8)


@composable.define_app
class arr2str:
    """convert array of uint8 to str"""

    def __init__(self, moltype: str = "dna", max_length: int = None):
        moltype = get_moltype(moltype)
        self.canonical = "".join(moltype)
        self.max_length = max_length
        extended = "".join(list(moltype.alphabets.degen))
        self.translation = b"".maketrans(
            "".join(chr(i) for i in range(len(extended))).encode("utf8"),
            extended.encode("utf8"),
        )

    def main(self, data: numpy.ndarray) -> str:
        if self.max_length:
            data = data[: self.max_length]

        b = data.tobytes().translate(self.translation)
        return bytearray(b).decode("utf8")


# we allow for file suffixes to include compression extensions
_fasta_format = re.compile("(fasta|mfa|faa|fna|fa)([.][a-zA-Z0-9]+)?$")
_genbank_format = re.compile("(genbank|gbk|gb|gbff)([.][a-zA-Z0-9]+)?$")


def get_seq_file_format(suffix: str) -> str:
    """returns format string

    Notes
    -----
    Based on cogent3 suffixes, returns either 'fasta' or 'genbank'
    or None.
    """
    if _fasta_format.match(suffix):
        return "fasta"
    return "genbank" if _genbank_format.match(suffix) else None


def determine_chunk_size(total_items, num_chunks):
    """chunk sizes for distributing items into approximately equal chunks"""
    base_chunk_size = total_items // num_chunks
    remainder = total_items % num_chunks

    return [
        base_chunk_size + 1 if i < remainder else base_chunk_size
        for i in range(num_chunks)
    ]


def chunked(iterable, num_chunks, verbose=False):
    sizes = determine_chunk_size(len(iterable), num_chunks)
    if verbose:
        print(f"chunk sizes: {sizes}")

    cum_sizes = numpy.array(sizes).cumsum()
    starts = [0] + cum_sizes[:-1].tolist()
    start_stop = numpy.array([starts, cum_sizes]).T
    for start, end in start_stop:
        yield iterable[start:end]


class summary_stats:
    """computes the summary statistics for a set of numbers"""

    def __init__(self, numbers: numpy.ndarray):
        self._numbers = numbers

    @functools.cached_property
    def n(self):
        return len(self._numbers)

    @functools.cached_property
    def mean(self):
        return math.fsum(self._numbers) / self.n

    @functools.cached_property
    def var(self):
        """unbiased estimate of the variance"""
        return math.fsum((x - self.mean) ** 2 for x in self._numbers) / (self.n - 1)

    @functools.cached_property
    def std(self):
        """standard deviation"""
        return self.var**0.5

    @functools.cached_property
    def cov(self):
        return self.std / self.mean


def _comma_sep_or_file(*args):
    include = args[-1]
    if include is None:
        return None
    if pathlib.Path(include).is_file():
        path = pathlib.Path(include)
        names = path.read_text().splitlines()
        return [name.strip() for name in names]
    return [n.strip() for n in include.split(",") if n.strip()]


class _printer:
    from rich.console import Console

    def __init__(self) -> None:
        self._console = self.Console()

    def __call__(self, txt: str, colour: str):
        """print text in colour"""
        msg = rich_text.Text(txt)
        msg.stylize(colour)
        self._console.print(msg)


print_colour = _printer()
